from dagster import Array, Field, Noneable, Shape, StringSource

SHARED_ECS_CONFIG = {
    "secrets": Field(
        Noneable(Array(Shape({"name": StringSource, "valueFrom": StringSource}))),
        is_required=False,
        description=(
            "An array of AWS Secrets Manager secrets. These secrets will "
            "be mounted as environment variabls in the container. See "
            "https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_Secret.html."
        ),
    ),
    "secrets_tags": Field(
        Noneable(Array(StringSource)),
        is_required=False,
        description=(
            "AWS Secrets Manager secrets with these tags will be mounted as "
            "environment variables in the container."
        ),
    ),
}



from typing import List

from dagster_cloud.api.dagster_cloud_api import DagsterCloudSandboxProxyInfo


def find_unallocated_sandbox_port(
    allocated_ports: List[int],
    proxy_info: DagsterCloudSandboxProxyInfo,
) -> str:
    """
    Multiple frpc clients can't be configured to run on the same port for the same
    frps server. Given a list of existing ports and the proxy information, find an
    unused port.
    """
    valid_ports = range(proxy_info.min_port, proxy_info.max_port + 1)

    try:
        return next(str(port) for port in valid_ports if port not in allocated_ports)
    except StopIteration:
        raise Exception("All available ports already in use.")



import logging
import os
import sys
import tempfile
import threading
import time
import zlib
from abc import abstractmethod
from contextlib import AbstractContextManager
from typing import Collection, Dict, Generic, List, NamedTuple, Optional, Set, TypeVar, Union

import dagster._check as check
from dagster.api.get_server_id import sync_get_server_id
from dagster.api.list_repositories import sync_list_repositories_grpc
from dagster.core.errors import DagsterUserCodeUnreachableError
from dagster.core.executor.step_delegating import StepHandler
from dagster.core.host_representation import ExternalRepositoryOrigin
from dagster.core.host_representation.grpc_server_registry import GrpcServerEndpoint
from dagster.core.host_representation.origin import (
    RegisteredRepositoryLocationOrigin,
    RepositoryLocationOrigin,
)
from dagster.core.instance import MayHaveInstanceWeakref
from dagster.core.launcher import RunLauncher
from dagster.grpc.client import DagsterGrpcClient
from dagster.grpc.types import GetCurrentImageResult
from dagster.serdes import deserialize_as, serialize_dagster_namedtuple, whitelist_for_serdes
from dagster.utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster_cloud.api.dagster_cloud_api import (
    DagsterCloudSandboxConnectionInfo,
    DagsterCloudSandboxProxyInfo,
    DagsterCloudUploadLocationData,
    DagsterCloudUploadRepositoryData,
    DagsterCloudUploadWorkspaceEntry,
)
from dagster_cloud.errors import raise_http_error
from dagster_cloud.execution.monitoring import (
    CloudRunWorkerStatuses,
    start_run_worker_monitoring_thread,
)
from dagster_cloud.instance import DagsterCloudAgentInstance
from dagster_cloud.util import diff_serializable_namedtuple_map
from dagster_cloud.workspace.origin import CodeDeploymentMetadata

DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT = 60


USER_CODE_LAUNCHER_RECONCILE_INTERVAL = 1

ServerHandle = TypeVar("ServerHandle")

SANDBOX_QUERY = """
    query DeploymentInfo {
         deploymentInfo {
             deploymentType
         }
     }
"""

AUTHORIZED_KEY_QUERY = """
    query DeploymentInfo {
         deploymentInfo {
             sshKeys {
                publicKey
            }
         }
     }
"""


@whitelist_for_serdes
class UserCodeLauncherEntry(
    NamedTuple(
        "_UserCodeLauncherEntry",
        [
            ("code_deployment_metadata", CodeDeploymentMetadata),
            ("update_timestamp", float),
            ("sandbox_saved_timestamp", Optional[float]),
            ("sandbox_proxy_info", Optional[DagsterCloudSandboxProxyInfo]),
        ],
    )
):
    def __new__(
        cls,
        code_deployment_metadata,
        update_timestamp,
        sandbox_saved_timestamp=None,
        sandbox_proxy_info=None,
    ):
        return super(UserCodeLauncherEntry, cls).__new__(
            cls,
            check.inst_param(
                code_deployment_metadata, "code_deployment_metadata", CodeDeploymentMetadata
            ),
            check.float_param(update_timestamp, "update_timestamp"),
            check.opt_float_param(sandbox_saved_timestamp, "sandbox_saved_timestamp"),
            check.opt_inst_param(
                sandbox_proxy_info, "sandbox_proxy_info", DagsterCloudSandboxProxyInfo
            ),
        )


class DagsterCloudUserCodeLauncher(
    AbstractContextManager, MayHaveInstanceWeakref[DagsterCloudAgentInstance], Generic[ServerHandle]
):
    def __init__(self):
        self._grpc_endpoints: Dict[str, Union[GrpcServerEndpoint, SerializableErrorInfo]] = {}
        self._grpc_endpoints_lock = threading.Lock()

        # periodically reconciles to make desired = actual
        self._desired_entries: Optional[Dict[str, UserCodeLauncherEntry]] = None
        self._actual_entries = {}
        self._metadata_lock = threading.Lock()

        self._upload_locations: Set[str] = set()

        self._logger = logging.getLogger("dagster_cloud")
        self._started: bool = False
        self._run_worker_monitoring_thread = None
        self._run_worker_monitoring_thread_shutdown_event = None
        self._run_worker_statuses_list = []
        self._run_worker_statuses_list_lock = threading.Lock()

        self._reconcile_count = 0
        self._reconcile_grpc_metadata_shutdown_event = threading.Event()
        self._reconcile_grpc_metadata_thread = None

        self._is_workspace_ready_lock = threading.Lock()
        self._is_workspace_ready: bool = False

        self._is_dev_sandbox = None
        super().__init__()

    @property
    def supports_dev_sandbox(self) -> bool:
        return False

    def is_dev_sandbox(self):
        if self._is_dev_sandbox == None and self.supports_dev_sandbox:
            client = self._instance.graphql_client
            result = client.execute(SANDBOX_QUERY)
            self._is_dev_sandbox = result["data"]["deploymentInfo"]["deploymentType"] == "DEV"
        return self._is_dev_sandbox

    def _authorized_key(self) -> Optional[str]:
        if self.is_dev_sandbox():
            client = self._instance.graphql_client
            result = client.execute(AUTHORIZED_KEY_QUERY)
            keys = [key["publicKey"] for key in result["data"]["deploymentInfo"]["sshKeys"]]
            if keys:
                # TODO: Support multiple keys
                return keys[0]
        return None

    def start(self, run_reconcile_thread=True):
        # Initialize
        self.is_dev_sandbox()

        check.invariant(
            not self._started,
            "Called start() on a DagsterCloudUserCodeLauncher that was already started",
        )
        # Begin spinning user code up and down
        self._started = True

        if self._instance.run_launcher.supports_check_run_worker_health:
            self._logger.debug("Starting run worker monitoring.")
            (
                self._run_worker_monitoring_thread,
                self._run_worker_monitoring_thread_shutdown_event,
            ) = start_run_worker_monitoring_thread(
                self._instance, self._run_worker_statuses_list, self._run_worker_statuses_list_lock
            )
        else:
            self._logger.debug(
                "Not starting run worker monitoring, because it's not supported on this agent."
            )

        self._cleanup_servers()

        if run_reconcile_thread:
            self._reconcile_grpc_metadata_thread = threading.Thread(
                target=self._reconcile_thread,
                args=(self._reconcile_grpc_metadata_shutdown_event,),
                name="grpc-reconcile-watch",
            )
            self._reconcile_grpc_metadata_thread.daemon = True
            self._reconcile_grpc_metadata_thread.start()

    def is_run_worker_monitoring_thread_alive(self):
        return (
            self._run_worker_monitoring_thread is not None
            and self._run_worker_monitoring_thread.is_alive()
        )

    def get_cloud_run_worker_statuses(self):
        if not self._instance.run_launcher.supports_check_run_worker_health:
            return CloudRunWorkerStatuses(
                None, run_worker_monitoring_supported=False, run_worker_monitoring_thread_alive=None
            )

        self._logger.debug("Getting cloud run worker statuses for a heartbeat")

        with self._run_worker_statuses_list_lock:
            # values are immutable, don't need deepcopy
            statuses_list = self._run_worker_statuses_list.copy()
        self._logger.debug("Returning statuses_list: {}".format(statuses_list))
        return CloudRunWorkerStatuses(
            statuses=statuses_list,
            run_worker_monitoring_supported=True,
            run_worker_monitoring_thread_alive=self.is_run_worker_monitoring_thread_alive(),
        )

    def supports_origin(self, repository_location_origin: RepositoryLocationOrigin) -> bool:
        return isinstance(repository_location_origin, RegisteredRepositoryLocationOrigin)

    @property
    def supports_reload(self) -> bool:
        return False

    def _update_workspace_entry(self, workspace_entry: DagsterCloudUploadWorkspaceEntry):
        with tempfile.TemporaryDirectory() as temp_dir:
            dst = os.path.join(temp_dir, "workspace_entry.tmp")
            with open(dst, "wb") as f:
                f.write(
                    zlib.compress(serialize_dagster_namedtuple(workspace_entry).encode("utf-8"))
                )

            with open(dst, "rb") as f:
                self._logger.info(
                    "Uploading workspace entry for {location_name} ({size} bytes)".format(
                        location_name=workspace_entry.location_name, size=os.path.getsize(dst)
                    )
                )

                connection_info = None
                if self.is_dev_sandbox():
                    connection_info = self.get_sandbox_connection_info(
                        workspace_entry.location_name,
                    )
                resp = self._instance.requests_session.post(
                    self._instance.dagster_cloud_upload_workspace_entry_url,
                    headers=self._instance.dagster_cloud_api_headers,
                    data={
                        "sandbox_connection_info": serialize_dagster_namedtuple(connection_info)
                        if connection_info
                        else None
                    },
                    files={"workspace_entry.tmp": f},
                    timeout=self._instance.dagster_cloud_api_timeout,
                )
                raise_http_error(resp)

                self._logger.info(
                    "Successfully uploaded workspace entry for {location_name}".format(
                        location_name=workspace_entry.location_name
                    )
                )

    def _get_upload_location_data(self, location_name: str) -> DagsterCloudUploadLocationData:
        location_origin = self._get_repository_location_origin(location_name)
        client = self.get_grpc_endpoint(location_origin).create_client()

        list_repositories_response = sync_list_repositories_grpc(client)

        upload_repo_datas: List[DagsterCloudUploadRepositoryData] = []

        for (
            repository_name,
            code_pointer,
        ) in list_repositories_response.repository_code_pointer_dict.items():
            external_repository_chunks = list(
                client.streaming_external_repository(
                    external_repository_origin=ExternalRepositoryOrigin(
                        location_origin,
                        repository_name,
                    )
                )
            )

            serialized_repository_data = "".join(
                [
                    chunk["serialized_external_repository_chunk"]
                    for chunk in external_repository_chunks
                ]
            )
            # Don't deserialize in case there are breaking changes - let the server do it
            upload_repo_datas.append(
                DagsterCloudUploadRepositoryData(
                    repository_name=repository_name,
                    code_pointer=code_pointer,
                    serialized_repository_data=serialized_repository_data,
                )
            )

        return DagsterCloudUploadLocationData(
            upload_repository_datas=upload_repo_datas,
            container_image=deserialize_as(
                client.get_current_image(), GetCurrentImageResult
            ).current_image,
            executable_path=list_repositories_response.executable_path,
        )

    def _update_location_error(
        self,
        location_name: str,
        error_info: SerializableErrorInfo,
        metadata: CodeDeploymentMetadata,
    ):
        self._logger.error(
            "Unable to load location {location_name}. Updating location with error data: {error_info}.".format(
                location_name=location_name,
                error_info=str(error_info),
            )
        )

        # Update serialized error
        errored_workspace_entry = DagsterCloudUploadWorkspaceEntry(
            location_name=location_name,
            deployment_metadata=metadata,
            upload_location_data=None,
            serialized_error_info=error_info,
        )

        self._update_workspace_entry(errored_workspace_entry)

    def _update_location_data(self, location_name: str, endpoint, metadata: CodeDeploymentMetadata):
        self._logger.info(
            "Uploading metadata for location {location_name}".format(location_name=location_name)
        )

        try:
            if isinstance(endpoint, SerializableErrorInfo):
                self._update_location_error(location_name, error_info=endpoint, metadata=metadata)
                return

            try:
                loaded_workspace_entry = DagsterCloudUploadWorkspaceEntry(
                    location_name=location_name,
                    deployment_metadata=metadata,
                    upload_location_data=self._get_upload_location_data(location_name),
                    serialized_error_info=None,
                )
            except Exception:
                self._update_location_error(
                    location_name,
                    error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                    metadata=metadata,
                )
                return

            self._logger.info(
                "Updating location {location_name} with repository load data".format(
                    location_name=location_name,
                )
            )
            self._update_workspace_entry(loaded_workspace_entry)
        except Exception:
            # Don't let a failure uploading the serialized data keep other locations
            # from being updated or continue reconciling in a loop
            error_info = serializable_error_info_from_exc_info(sys.exc_info())
            self._logger.error(
                "Error while writing location data for location {location_name}: {error_info}".format(
                    location_name=location_name,
                    error_info=error_info,
                )
            )

    @property
    def is_workspace_ready(self) -> bool:
        with self._is_workspace_ready_lock:
            return self._is_workspace_ready

    @property
    @abstractmethod
    def requires_images(self) -> bool:
        pass

    @abstractmethod
    def _get_server_handles_for_location(self, location_name: str) -> Collection[ServerHandle]:
        """Return a list of 'handles' that represent all running servers for a given location.
        Typically this will be a single server (unless an error was previous raised during a
        reconciliation loop. ServerHandle can be any type that is sufficient for
        _remove_server_handle to remove the service."""

    @abstractmethod
    def _create_new_server_endpoint(
        self, location_name: str, metadata: CodeDeploymentMetadata
    ) -> GrpcServerEndpoint:
        """Create a new server for the given location using the given metadata as configuration
        and return a GrpcServerEndpoint indicating a hostname/port that can be used to access
        the server. Should result in an additional handle being returned from _get_server_handles_for_location."""

    @abstractmethod
    def _create_dev_sandbox_endpoint(
        self,
        location_name: str,
        metadata: CodeDeploymentMetadata,
        authorized_key: str,
        proxy_info: DagsterCloudSandboxProxyInfo,
    ) -> GrpcServerEndpoint:
        """Create a new dev sandbox for the given location using the given metadata as configuration
        and return a GrpcServerEndpoint indicating a hostname/port that can be used to access
        the server. Should result in an additional handle being returned from _get_server_handles_for_location."""

    @abstractmethod
    def get_sandbox_connection_info(self, location_name: str) -> DagsterCloudSandboxConnectionInfo:
        pass

    @abstractmethod
    def _remove_server_handle(self, server_handle: ServerHandle) -> None:
        """Shut down any resources associated with the given handle. Called both during updates
        to spin down the old server once a new server has been spun up, and during removal."""

    @abstractmethod
    def _cleanup_servers(self):
        """Remove all servers, across all locations."""

    @abstractmethod
    def get_step_handler(self, execution_config: Optional[Dict]) -> StepHandler:
        pass

    @abstractmethod
    def run_launcher(self) -> RunLauncher:
        pass

    def __exit__(self, exception_type, exception_value, traceback):
        if self._reconcile_grpc_metadata_thread:
            self._reconcile_grpc_metadata_shutdown_event.set()
            self._reconcile_grpc_metadata_thread.join()

        if self._run_worker_monitoring_thread:
            self._run_worker_monitoring_thread_shutdown_event.set()
            self._run_worker_monitoring_thread.join()

        if self._started:
            self._cleanup_servers()

        super().__exit__(exception_value, exception_value, traceback)

    def update_grpc_metadata(
        self,
        desired_metadata: Dict[str, UserCodeLauncherEntry],
        upload_locations: Set[str],
    ):
        check.dict_param(
            desired_metadata,
            "desired_metadata",
            key_type=str,
            value_type=UserCodeLauncherEntry,
        )
        check.set_param(upload_locations, "upload_locations", str)

        with self._metadata_lock:
            # Change the inputs for the next reconciliation loop
            self._desired_entries = desired_metadata
            self._upload_locations = self._upload_locations.union(upload_locations)

    def _get_repository_location_origin(
        self, location_name: str
    ) -> RegisteredRepositoryLocationOrigin:
        return RegisteredRepositoryLocationOrigin(location_name)

    def _reconcile_thread(self, shutdown_event):
        while True:
            shutdown_event.wait(USER_CODE_LAUNCHER_RECONCILE_INTERVAL)
            if shutdown_event.is_set():
                break

            try:
                self.reconcile()
            except Exception:
                self._logger.error(
                    "Failure updating user code servers: {exc_info}".format(
                        exc_info=serializable_error_info_from_exc_info(sys.exc_info()),
                    )
                )

    def reconcile(self):
        with self._metadata_lock:
            desired_entries = (
                self._desired_entries.copy() if self._desired_entries != None else None
            )
            upload_locations = self._upload_locations.copy()
            self._upload_locations = set()

        if desired_entries == None:
            # Wait for the first time the desired metadata is set before reconciling
            return

        self._reconcile(desired_entries, upload_locations)
        self._reconcile_count += 1

        if self._reconcile_count == 1:
            with self._is_workspace_ready_lock:
                self._is_workspace_ready = True

    def _check_for_image(self, metadata: CodeDeploymentMetadata):
        if self.requires_images and not metadata.image:
            raise Exception(
                "Your agent's configuration requires you to specify an image. "
                "Use the `--image` flag when specifying your location to tell the agent "
                "which image to use to load your code."
            )

        if (not self.requires_images) and metadata.image:
            raise Exception(
                "Your agent's configuration cannot load locations that specify a Docker "
                "image. Either update your location to not include an image, or change the `user_code_launcher` "
                "field in your agent's `dagster.yaml` file to a launcher that can load Docker images. "
            )

    def _reconcile(self, desired_entries, upload_locations):
        diff = diff_serializable_namedtuple_map(desired_entries, self._actual_entries)

        to_update_keys = diff.to_add.union(diff.to_update)
        existing_server_handles: Dict[str, List[ServerHandle]] = {}
        for to_update_key in to_update_keys:
            try:

                code_deployment_metadata = desired_entries[to_update_key].code_deployment_metadata

                self._logger.info(
                    "Updating server for location {location_name}".format(
                        location_name=to_update_key
                    )
                )
                existing_server_handles[to_update_key] = self._get_server_handles_for_location(
                    to_update_key
                )

                self._check_for_image(code_deployment_metadata)

                if self.is_dev_sandbox():
                    if self._should_soft_reload(
                        location_name=to_update_key,
                        entry=desired_entries[to_update_key],
                    ):
                        new_updated_endpoint = self._soft_reload(to_update_key)
                    else:
                        new_updated_endpoint = self._create_dev_sandbox_endpoint(
                            to_update_key,
                            code_deployment_metadata,
                            authorized_key=self._authorized_key(),
                            proxy_info=desired_entries[to_update_key].sandbox_proxy_info,
                        )
                else:
                    new_updated_endpoint = self._create_new_server_endpoint(
                        to_update_key,
                        code_deployment_metadata,
                    )

                self._logger.info(
                    "Created a new server for location {location_name}".format(
                        location_name=to_update_key
                    )
                )
            except Exception:
                error_info = serializable_error_info_from_exc_info(sys.exc_info())
                self._logger.error(
                    "Error while updating server for {to_update_key}: {error_info}".format(
                        to_update_key=to_update_key,
                        error_info=error_info,
                    )
                )
                new_updated_endpoint: SerializableErrorInfo = error_info

            with self._grpc_endpoints_lock:
                self._grpc_endpoints[to_update_key] = new_updated_endpoint

            if to_update_key in upload_locations:
                upload_locations.remove(to_update_key)
                self._update_location_data(
                    to_update_key,
                    new_updated_endpoint,
                    desired_entries[to_update_key].code_deployment_metadata,
                )

        for to_update_key in to_update_keys:
            # If a hard reload, clean up the server handles
            # If a soft reload, don't because we changed them in place
            if not self._should_soft_reload(
                location_name=to_update_key,
                entry=desired_entries[to_update_key],
            ):
                server_handles = existing_server_handles.get(to_update_key, [])

                if server_handles:
                    self._logger.info(
                        "Removing {num_servers} existing servers for location {location_name}".format(
                            num_servers=len(server_handles), location_name=to_update_key
                        )
                    )

                for server_handle in server_handles:
                    try:
                        self._remove_server_handle(server_handle)
                    except Exception:
                        self._logger.error(
                            "Error while cleaning up after updating server for {to_update_key}: {error_info}".format(
                                to_update_key=to_update_key,
                                error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                            )
                        )

                if server_handles:
                    self._logger.info(
                        "Removed all previous servers for {location_name}".format(
                            location_name=to_update_key
                        )
                    )

            # Always update our actual entries
            self._actual_entries[to_update_key] = desired_entries[to_update_key]

        for to_remove_key in diff.to_remove:
            try:
                self._remove_server(to_remove_key)
            except Exception:
                self._logger.error(
                    "Error while removing server for {to_remove_key}: {error_info}".format(
                        to_remove_key=to_remove_key,
                        error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                    )
                )

            with self._grpc_endpoints_lock:
                del self._grpc_endpoints[to_remove_key]
            del self._actual_entries[to_remove_key]

        # Upload any locations that were requested to be uploaded, but weren't updated
        # as part of this reconciliation loop
        for location in upload_locations:
            with self._grpc_endpoints_lock:
                endpoint = self._grpc_endpoints.get(location)

            self._update_location_data(
                location,
                endpoint,
                self._actual_entries[location].code_deployment_metadata,
            )

    def get_grpc_endpoint(
        self, repository_location_origin: RepositoryLocationOrigin
    ) -> GrpcServerEndpoint:
        with self._grpc_endpoints_lock:
            location_name = repository_location_origin.location_name
            endpoint = self._grpc_endpoints.get(location_name)

        if not endpoint:
            raise DagsterUserCodeUnreachableError(
                f"No server endpoint exists for location {location_name}"
            )

        if isinstance(endpoint, SerializableErrorInfo):
            # Consider raising the original exception here instead of a wrapped one
            raise DagsterUserCodeUnreachableError(
                f"Failure loading server endpoint for location {location_name}: {endpoint}"
            )

        return endpoint

    def get_grpc_endpoints(
        self,
    ) -> Dict[str, Union[GrpcServerEndpoint, SerializableErrorInfo]]:
        with self._grpc_endpoints_lock:
            return self._grpc_endpoints.copy()

    def _remove_server(self, location_name: str):
        self._logger.info(
            "Removing server for location {location_name}".format(location_name=location_name)
        )
        existing_server_handles = self._get_server_handles_for_location(location_name)
        for server_handle in existing_server_handles:
            self._remove_server_handle(server_handle)

    def _wait_for_server(self, host: str, port: int, timeout, socket: Optional[str] = None) -> str:
        # Wait for the server to be ready (while also loading the server ID)
        server_id = None
        start_time = time.time()

        last_error = None

        while True:
            client = DagsterGrpcClient(port=port, host=host, socket=socket)
            try:
                server_id = sync_get_server_id(client)
                break
            except Exception:
                last_error = serializable_error_info_from_exc_info(sys.exc_info())

            if time.time() - start_time > timeout:
                raise Exception(
                    f"Timed out waiting for server {host}:{port}. Most recent connection error: {str(last_error)}"
                )

            time.sleep(1)
        return server_id

    def _should_soft_reload(self, location_name: str, entry: UserCodeLauncherEntry) -> bool:
        if not self.is_dev_sandbox():
            return False

        existing_entry = self._actual_entries.get(location_name)
        if not existing_entry:
            return False

        # A hard reload has been triggered
        if entry.update_timestamp > existing_entry.update_timestamp:
            return False

        if not existing_entry.sandbox_saved_timestamp:
            return True

        if entry.sandbox_saved_timestamp > existing_entry.sandbox_saved_timestamp:
            return True

        # Getting here shouldn't be possible
        return False

    def _soft_reload(self, to_update_key):
        existing_dev_sandbox_endpoint = self.get_grpc_endpoints().get(to_update_key)
        if existing_dev_sandbox_endpoint:
            server_id = existing_dev_sandbox_endpoint.server_id
            host = existing_dev_sandbox_endpoint.host
            port = existing_dev_sandbox_endpoint.port

            client = existing_dev_sandbox_endpoint.create_client()
            client.shutdown_server()

            start_time = time.time()
            last_error = f"Server {server_id} is still running."
            while True:
                try:
                    # Wait for the server ID to change; this means we've reloaded the gRPC server.
                    if server_id != sync_get_server_id(client):
                        break
                except Exception:
                    last_error = serializable_error_info_from_exc_info(sys.exc_info())
                if time.time() - start_time > 15:
                    raise Exception(
                        f"Timed out waiting for server {host}:{port} to reload. Most recent error: {str(last_error)}"
                    )
                time.sleep(1)
            new_updated_endpoint = GrpcServerEndpoint(
                server_id=sync_get_server_id(client),
                host=host,
                port=port,
                socket=None,
            )
            return new_updated_endpoint



# pylint: disable=unused-argument

import os
import re
import subprocess
from collections import Counter
from datetime import timedelta
from typing import Dict, List, Optional

import typer
import yaml
from dagster_cloud.api.dagster_cloud_api import DagsterCloudSandboxConnectionInfo
from dagster_cloud.cli.config_utils import dagster_cloud_options
from dagster_cloud.cli.sandbox.utils import get_current_display_timestamp
from dagster_cloud.storage.client import GqlShimClient
from typer import Typer

from .. import gql, ui
from .sync_method import MutagenSyncMethod, SyncMethod, SyncState, SyncedDirectory

CONNECTION_INFO_GIVE_UP_TIME = timedelta(seconds=120)

app = Typer(help="Interface with your dev sandbox.")


def get_all_connection_info(client: GqlShimClient) -> Dict[str, DagsterCloudSandboxConnectionInfo]:
    """
    Returns a mapping of location name to sandbox connection info for all locations.
    """
    workspace_entries = gql.fetch_workspace_entries(client)
    return {
        workspace_entry["locationName"]: DagsterCloudSandboxConnectionInfo(
            username=workspace_entry["connectionInfo"]["username"],
            hostname=workspace_entry["connectionInfo"]["hostname"],
            port=workspace_entry["connectionInfo"]["port"],
        )
        for workspace_entry in workspace_entries
        if workspace_entry.get("connectionInfo")
    }


def launch_ssh(arguments: List[str]) -> None:
    os.execvp("ssh", ["ssh", *arguments])


def get_default_sync_method() -> SyncMethod:
    return MutagenSyncMethod()


def dev_deployment_client_or_error(
    base_client_url: str, organization: str, api_token: str
) -> GqlShimClient:
    default_client = gql.graphql_client_from_url(base_client_url, api_token)
    try:
        dev_depl = gql.get_dev_deployment_name(default_client)
    except Exception:
        raise ui.error(
            "No dev sandbox found for your user - you need to set one up before you can use this command."
        )
    # Current deployment is dev depl
    if not dev_depl:
        return default_client

    # Else, construct new gql client
    return gql.graphql_client_from_url(gql.url_from_config(organization, dev_depl), api_token)


@app.command(name="ssh")
@dagster_cloud_options(requires_url=True)
def ssh(
    organization: str,
    deployment: str,
    api_token: str,
    url: str,
    location_name=typer.Argument(
        ..., help="The name of the code location whose user code server to connect to."
    ),
) -> None:
    """
    Connects to the user code server hosting a given code location.
    """
    client = dev_deployment_client_or_error(url, organization, api_token)

    all_connection_info = get_all_connection_info(client)

    if not location_name in all_connection_info:
        raise ui.error(f"Cannot find sandbox user code server for location {location_name}.")

    connection_info = all_connection_info[location_name]

    # TODO forward all other arguments to SSH
    launch_ssh(
        ["-p", str(connection_info.port), f"{connection_info.username}@{connection_info.hostname}"]
    )


def find_locations_yaml(search_directory: str) -> Optional[str]:
    """
    Given a search directory, finds a locations.yaml file located
    in the search directory or a parent. Returns None if none is found.
    """
    curr_dir = search_directory

    while not os.path.exists(os.path.join(curr_dir, "locations.yaml")):
        parent_dir = os.path.dirname(curr_dir)
        if parent_dir == curr_dir:
            return None
        curr_dir = parent_dir
    return curr_dir


def get_directory_identifier(location_name: str, index: int) -> str:
    return f"{re.sub('[^0-9a-zA-Z-]+', '-', location_name)}-{index}"


def get_directories_to_sync(
    directory: str, all_connection_info: Dict[str, DagsterCloudSandboxConnectionInfo]
) -> List[SyncedDirectory]:
    """
    Finds a set of code sync mappings defined in a locations.yaml file,
    combines with the sandbox connection info to create a list of directories to sync.
    """
    locations_yaml_dir = find_locations_yaml(directory)
    if not locations_yaml_dir:
        raise ValueError("No locations.yaml file found")

    directories_to_sync: List[SyncedDirectory] = []
    with open(os.path.join(locations_yaml_dir, "locations.yaml"), encoding="utf8") as f:
        locations = yaml.safe_load(f.read())["locations"]

        defined_locations = set(locations.keys())
        available_locations = set(all_connection_info.keys())

        if len(available_locations.intersection(defined_locations)) == 0:
            defined_locations_display = ", ".join(locations.keys()) or "(None)"
            available_locations_display = ", ".join(all_connection_info.keys()) or "(None)"
            raise ui.error(
                "No accessable sync locations found\n\n"
                f"Defined in locations.yaml: {defined_locations_display}\nAvailable code locations: {available_locations_display}"
            )

        if len(defined_locations - available_locations) > 0:
            not_synced = ", ".join(defined_locations - available_locations)
            ui.warn(
                f"Some locations defined in locations.yaml are not accessible to sync: {not_synced}\n"
            )

        if len(available_locations - defined_locations) > 0:
            not_synced = ", ".join(available_locations - defined_locations)
            ui.warn(
                f"Some available locations are not defined in locations.yaml, and will not be synced: {not_synced}\n"
            )

        for location_name, location in locations.items():
            if location_name in all_connection_info:
                connection_info = all_connection_info[location_name]
                for i, mapping in enumerate(location.get("code_sync", [])):
                    directories_to_sync.append(
                        SyncedDirectory(
                            identifier=get_directory_identifier(location_name, i),
                            location_name=location_name,
                            from_directory=str(
                                os.path.join(locations_yaml_dir, mapping.get("from"))
                            ),
                            to_directory=mapping.get("to"),
                            connection_info=connection_info,
                        )
                    )
    return directories_to_sync


def validate_host_connection(connection_info: DagsterCloudSandboxConnectionInfo):
    """
    Validates that we can SSH into the given host, and prompts the user to add to
    known_hosts if need be.
    """
    subprocess.check_output(
        [
            "ssh",
            "-p",
            f"{connection_info.port}",
            "-o",
            "ConnectTimeout=10",
            f"{connection_info.username}@{connection_info.hostname}",
            "exit",
        ]
    )


def print_sync_state(state_map: Dict[str, SyncState], remove_previous: bool = True):
    """
    Given a mapping from directory identifiers to sync states, prints a summary of
    all sync states.
    """
    if remove_previous:
        ui.erase_previous_line(2)

    c = Counter(state_map.values())
    n_unavailable = c.get(SyncState.UNAVAILABLE, 0)
    n_syncing = c.get(SyncState.SYNCING, 0)
    n_synced = c.get(SyncState.SYNCED, 0)
    total = len(state_map.keys()) - n_unavailable

    if n_syncing > 0:
        ui.print(
            ui.yellow(
                f"[{get_current_display_timestamp()}] ({n_synced}/{total}) Syncing directories..."
            )
        )
    else:
        ui.print(ui.blue(f"[{get_current_display_timestamp()}] ({n_synced}/{total}) Sync complete"))

    if n_unavailable > 0:
        ui.print(ui.red(f"           {n_unavailable} directories unable to sync"))
    elif remove_previous:
        ui.print()


@app.command(name="sync")
@dagster_cloud_options(requires_url=True)
def code_sync(
    organization: str,
    deployment: str,
    api_token: str,
    url: str,
    verbose: bool = typer.Option(False, "--verbose", "-v", is_flag=True),
    very_verbose: bool = typer.Option(False, "--very-verbose", "-vv", is_flag=True),
):
    """
    Starts to synchronize code in the local environment with code in your cloud sandbox.
    """
    if verbose and very_verbose:
        raise ui.error(
            f"Cannot specify both {ui.as_code('--verbose')} and {ui.as_code('--very-verbose')} options"
        )
    verbosity_level = 2 if very_verbose else 1 if verbose else 0

    client = dev_deployment_client_or_error(url, organization, api_token)

    sync_method = get_default_sync_method()
    sync_method.preflight(verbosity_level)

    # Populate list of locations to sync
    all_connection_info = get_all_connection_info(client)
    for connection_info in all_connection_info.values():
        try:
            validate_host_connection(connection_info)
        except Exception:
            raise ui.error("Unable to establish connection")

    directories_to_sync = get_directories_to_sync(os.path.abspath(os.getcwd()), all_connection_info)

    ui.print("Enabling code sync...")

    for directory in directories_to_sync:
        sync_method.create_directory_sync(directory)

    identifier_location_name_map = {
        directory.identifier: directory.location_name for directory in directories_to_sync
    }
    try:
        states_map = {dir.identifier: SyncState.SYNCING for dir in directories_to_sync}

        print()
        print_sync_state(states_map, remove_previous=True)

        for identifier, new_state in sync_method.sync_loop():
            last_state = states_map[identifier]
            states_map[identifier] = new_state

            if new_state == SyncState.SYNCED and last_state != new_state:
                gql.reload_repo_location(client, identifier_location_name_map[identifier])

            if (
                new_state == SyncState.SYNCING
                or new_state == SyncState.UNAVAILABLE
                or last_state != new_state
            ):
                print_sync_state(states_map, remove_previous=verbosity_level == 0)

    except KeyboardInterrupt:
        pass
    finally:
        ui.print("Terminating sync...")
        for directory in directories_to_sync:
            sync_method.cleanup_directory_sync(directory.identifier)



import platform
from typing import Dict, Optional

from dagster.utils import merge_dicts

from ..auth import API_TOKEN_HEADER, DEPLOYMENT_NAME_HEADER
from ..version import __version__
from .versioning.constants import DAGSTER_CLOUD_VERSION_HEADER, PYTHON_VERSION_HEADER


def get_dagster_cloud_api_headers(
    agent_token: str,
    deployment_name: Optional[str] = None,
    additional_headers: Optional[Dict[str, str]] = None,
) -> Dict[str, str]:
    return merge_dicts(
        {
            API_TOKEN_HEADER: agent_token,
            PYTHON_VERSION_HEADER: platform.python_version(),
            DAGSTER_CLOUD_VERSION_HEADER: __version__,
        },
        {DEPLOYMENT_NAME_HEADER: deployment_name} if deployment_name else {},
        additional_headers if additional_headers else {},
    )



from enum import Enum

from dagster._check import not_implemented


class TaskStatus(Enum):
    """
    Enum for the status of a task.
    """

    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    NOT_FOUND = "NOT_FOUND"
    NOT_IMPLEMENTED = "NOT_IMPLEMENTED"
    UNKNOWN = "UNKNOWN"



import logging
import os
import sys
import time
import uuid
from typing import Any, Collection, Dict, List, Optional

import docker
from dagster import Array, Field, IntSource, Permissive
from dagster import _check as check
from dagster.core.host_representation.grpc_server_registry import GrpcServerEndpoint
from dagster.serdes import ConfigurableClass
from dagster.utils import find_free_port, merge_dicts
from dagster_cloud.api.dagster_cloud_api import (
    DagsterCloudSandboxConnectionInfo,
    DagsterCloudSandboxProxyInfo,
)
from dagster_cloud.execution.step_handler.docker_step_handler import DockerStepHandler
from dagster_cloud.workspace.origin import CodeDeploymentMetadata
from dagster_docker import DockerRunLauncher
from dagster_docker.container_context import DockerContainerContext
from dagster_docker.utils import parse_env_var
from docker.models.containers import Container

from ..config_schema.docker import SHARED_DOCKER_CONFIG
from ..user_code_launcher import (
    DAGSTER_SANDBOX_PORT_ENV,
    DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
    DagsterCloudUserCodeLauncher,
    find_unallocated_sandbox_port,
)

GRPC_SERVER_LABEL = "dagster_grpc_server"

IMAGE_PULL_LOG_INTERVAL = 15


class DockerUserCodeLauncher(DagsterCloudUserCodeLauncher[Container], ConfigurableClass):
    def __init__(
        self,
        inst_data=None,
        networks=None,
        env_vars=None,
        container_kwargs=None,
        server_process_startup_timeout=None,
    ):
        self._inst_data = inst_data
        self._logger = logging.getLogger("dagster_cloud")
        self._networks = check.opt_list_param(networks, "networks", of_type=str)
        self._input_env_vars = check.opt_list_param(env_vars, "env_vars", of_type=str)

        self._container_kwargs = check.opt_dict_param(
            container_kwargs, "container_kwargs", key_type=str
        )

        self._server_process_startup_timeout = check.opt_int_param(
            server_process_startup_timeout,
            "server_process_startup_timeout",
            DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
        )

        super(DockerUserCodeLauncher, self).__init__()

    @property
    def requires_images(self):
        return True

    @property
    def inst_data(self):
        return self._inst_data

    @property
    def env_vars(self) -> List[str]:
        return self._input_env_vars + self._instance.dagster_cloud_api_env_vars

    @property
    def container_kwargs(self) -> Dict[str, Any]:
        return self._container_kwargs

    @classmethod
    def config_type(cls):
        return merge_dicts(
            SHARED_DOCKER_CONFIG,
            {
                "server_process_startup_timeout": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
                    description="Timeout when waiting for a code server to be ready after it is created",
                ),
            },
        )

    @staticmethod
    def from_config_value(inst_data, config_value):
        return DockerUserCodeLauncher(inst_data=inst_data, **config_value)

    def _create_container(
        self,
        client,
        location_name,
        metadata,
        container_name,
        hostname,
        environment,
        ports,
        container_context,
        command,
    ):
        return client.containers.create(
            metadata.image,
            detach=True,
            hostname=hostname,
            name=container_name,
            network=container_context.networks[0] if len(container_context.networks) else None,
            environment=environment,
            labels=[GRPC_SERVER_LABEL, self._get_label_for_location(location_name)],
            command=command,
            ports=ports,
            **container_context.container_kwargs,
        )

    def _get_server_handles_for_location(self, location_name: str) -> Collection[Container]:
        client = docker.client.from_env()
        return client.containers.list(
            all=True, filters={"label": self._get_label_for_location(location_name)}
        )

    @property
    def supports_dev_sandbox(self) -> bool:
        return True

    def _create_dev_sandbox_endpoint(
        self,
        location_name: str,
        metadata: CodeDeploymentMetadata,
        authorized_key: str,
        proxy_info: DagsterCloudSandboxProxyInfo,
    ) -> GrpcServerEndpoint:
        port = find_unallocated_sandbox_port(
            allocated_ports=list_allocated_sandbox_ports(),
            proxy_info=proxy_info,
        )

        return self._launch(
            location_name=location_name,
            metadata=metadata,
            command=["supervisord"],
            additional_environment={  # TODO: Abstract into SandboxContainerEnvironment
                "DAGSTER_SANDBOX_AUTHORIZED_KEY": authorized_key,
                "DAGSTER_PROXY_HOSTNAME": proxy_info.hostname,
                "DAGSTER_PROXY_PORT": str(proxy_info.port),
                "DAGSTER_PROXY_AUTH_TOKEN": proxy_info.auth_token,
                DAGSTER_SANDBOX_PORT_ENV: port,
            },
        )

    def get_sandbox_connection_info(self, location_name: str) -> DagsterCloudSandboxConnectionInfo:
        # TODO: Re-implement this in the base class. We'll need to extend handles to include
        # info like the environment and created timestamp or add abstract methods to look
        # this information up.

        # If there are multiple handles for a location,
        # get the most recently created one. We assume this
        # is the "new" one in our blue/green deployment.
        containers = sorted(
            list(self._get_server_handles_for_location(location_name)),
            key=lambda handle: handle.attrs["Created"],
        )

        try:
            container = containers[-1]
        except IndexError:
            raise Exception(f"{location_name} has no running containers.")

        port_env = next(
            (env for env in container.attrs["Config"]["Env"] if "DAGSTER_SANDBOX_PORT" in env)
        )
        port = port_env.split("=")[1]

        hostname_env = next(
            (env for env in container.attrs["Config"]["Env"] if "DAGSTER_PROXY_HOSTNAME" in env)
        )
        hostname = hostname_env.split("=")[1]

        username = "root"  # Hardcoded for now

        return DagsterCloudSandboxConnectionInfo(
            username=username,
            hostname=hostname,
            port=int(port),
        )

    def _create_new_server_endpoint(
        self, location_name: str, metadata: CodeDeploymentMetadata
    ) -> GrpcServerEndpoint:
        return self._launch(
            location_name=location_name,
            metadata=metadata,
            command=metadata.get_grpc_server_command(),
        )

    def _launch(
        self,
        location_name: str,
        metadata: CodeDeploymentMetadata,
        command: List[str],
        additional_environment: Optional[Dict[str, str]] = None,
    ) -> GrpcServerEndpoint:
        client = docker.client.from_env()

        container_context = DockerContainerContext(
            registry=None,
            env_vars=self.env_vars,
            networks=self._networks,
            container_kwargs=self._container_kwargs,
        ).merge(DockerContainerContext.create_from_config(metadata.container_context))

        container_name = f"{location_name}_{str(uuid.uuid4().hex)[0:6]}"

        self._logger.info(
            "Starting a new container for location {location_name} with image {image}: {container_name}".format(
                location_name=location_name, image=metadata.image, container_name=container_name
            )
        )

        ports = {}

        has_network = len(self._networks) > 0
        if has_network:
            grpc_port = 4000
            hostname = container_name
        else:
            grpc_port = find_free_port()
            ports[grpc_port] = grpc_port
            hostname = "localhost"

        environment = merge_dicts(
            (dict([parse_env_var(env_var) for env_var in container_context.env_vars])),
            metadata.get_grpc_server_env(grpc_port),
            additional_environment or {},
            {"DAGSTER_SERVER_NAME": container_name},
        )

        try:
            container = self._create_container(
                client,
                location_name,
                metadata,
                container_name,
                hostname,
                environment,
                ports,
                container_context,
                command,
            )
        except docker.errors.ImageNotFound:
            last_log_time = time.time()
            self._logger.info("Pulling image {image}...".format(image=metadata.image))
            for _line in docker.APIClient().pull(metadata.image, stream=True):
                if time.time() - last_log_time > IMAGE_PULL_LOG_INTERVAL:
                    self._logger.info("Still pulling image {image}...".format(image=metadata.image))
                    last_log_time = time.time()

            container = self._create_container(
                client,
                location_name,
                metadata,
                container_name,
                hostname,
                environment,
                ports,
                container_context,
                command,
            )

        if len(container_context.networks) > 1:
            for network_name in container_context.networks[1:]:
                network = client.networks.get(network_name)
                network.connect(container)

        container.start()

        self._logger.info("Started container {container_id}".format(container_id=container.id))

        server_id = self._wait_for_server(
            host=hostname, port=grpc_port, timeout=self._server_process_startup_timeout
        )

        endpoint = GrpcServerEndpoint(
            server_id=server_id,
            host=hostname,
            port=grpc_port,
            socket=None,
        )

        return endpoint

    def _get_label_for_location(self, location_name):
        return f"location_{location_name}"

    def _remove_server_handle(self, server_handle: Container) -> None:
        container = server_handle
        try:
            container.stop()
        except Exception:
            self._logger.error(
                "Failure stopping container {container_id}: {exc_info}".format(
                    container_id=container.id,
                    exc_info=sys.exc_info(),
                )
            )
        container.remove(force=True)
        self._logger.info("Removed container {container_id}".format(container_id=container.id))

    def _cleanup_servers(self) -> None:
        client = docker.client.from_env()

        containers = client.containers.list(all=True, filters={"label": GRPC_SERVER_LABEL})
        for container in containers:
            self._remove_server_handle(container)

    def get_step_handler(self, _execution_config: Optional[Dict]) -> DockerStepHandler:
        return DockerStepHandler(self._networks, self.env_vars, self._container_kwargs)

    def run_launcher(self):
        launcher = DockerRunLauncher(
            image=None,
            env_vars=self.env_vars,
            networks=self._networks,
            container_kwargs=self._container_kwargs,
        )
        launcher.register_instance(self._instance)

        return launcher


def list_allocated_sandbox_ports() -> List[int]:
    client = docker.client.from_env()
    containers = client.containers.list(
        filters={
            "label": GRPC_SERVER_LABEL,
        },
    )

    allocated_ports = []
    for container in containers:
        env_list = container.attrs["Config"]["Env"]
        env = dict(item.split("=", 1) for item in env_list)
        port = env.get("DAGSTER_SANDBOX_PORT")
        if port:
            allocated_ports.append(int(port))

    return sorted(allocated_ports)



import logging
import sys
import threading
from collections import namedtuple
from typing import List, NamedTuple, Optional

import dagster._check as check
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.launcher import CheckRunHealthResult, WorkerStatus
from dagster.core.storage.pipeline_run import IN_PROGRESS_RUN_STATUSES, PipelineRunsFilter
from dagster.serdes import whitelist_for_serdes
from dagster.utils.error import serializable_error_info_from_exc_info
from dagster_cloud.instance import DagsterCloudAgentInstance, InstanceRef


@whitelist_for_serdes
class CloudRunWorkerStatus(
    NamedTuple(
        "_CloudRunWorkerStatus",
        [("run_id", str), ("status_type", WorkerStatus), ("message", Optional[str])],
    )
):
    def __new__(cls, run_id: str, status_type: WorkerStatus, message: Optional[str] = None):
        return super(CloudRunWorkerStatus, cls).__new__(
            cls,
            run_id=check.str_param(run_id, "run_id"),
            status_type=check.inst_param(status_type, "status_type", WorkerStatus),
            message=check.opt_str_param(message, "message"),
        )

    @classmethod
    def from_check_run_health_result(cls, run_id: str, result: CheckRunHealthResult):
        check.inst_param(result, "result", CheckRunHealthResult)
        return CloudRunWorkerStatus(run_id, result.status, result.msg)


@whitelist_for_serdes
class CloudRunWorkerStatuses(
    NamedTuple(
        "_CloudRunWorkerStatuses",
        [
            ("statuses", List[CloudRunWorkerStatus]),
            ("run_worker_monitoring_supported", bool),
            ("run_worker_monitoring_thread_alive", Optional[bool]),
        ],
    )
):
    def __new__(
        cls,
        statuses: Optional[List[CloudRunWorkerStatus]],
        run_worker_monitoring_supported: bool,
        run_worker_monitoring_thread_alive: Optional[bool],
    ):
        return super(CloudRunWorkerStatuses, cls).__new__(
            cls,
            statuses=check.opt_list_param(statuses, "statuses", of_type=CloudRunWorkerStatus),
            run_worker_monitoring_supported=check.bool_param(
                run_worker_monitoring_supported, "run_worker_monitoring_supported"
            ),
            run_worker_monitoring_thread_alive=check.opt_bool_param(
                run_worker_monitoring_thread_alive, "run_worker_monitoring_thread_alive"
            ),
        )


def get_cloud_run_worker_statuses(instance):
    launcher = instance.run_launcher
    check.invariant(launcher.supports_check_run_worker_health)
    runs = instance.get_runs(PipelineRunsFilter(statuses=IN_PROGRESS_RUN_STATUSES))
    return [
        CloudRunWorkerStatus.from_check_run_health_result(
            run.run_id, launcher.check_run_worker_health(run)
        )
        for run in runs
    ]


def run_worker_monitoring_thread(
    instance_ref: InstanceRef,
    statuses_list: List[CloudRunWorkerStatus],
    statuses_lock: threading.Lock,
    shutdown_event: threading.Event,
):
    logger = logging.getLogger("dagster_cloud")
    check.inst_param(instance_ref, "instance_ref", InstanceRef)
    logger.debug("Run Monitor thread has started")
    with DagsterCloudAgentInstance.from_ref(instance_ref) as instance:
        while not shutdown_event.is_set():
            run_worker_monitoring_thread_iteration(instance, statuses_list, statuses_lock, logger)
            shutdown_event.wait(instance.dagster_cloud_run_worker_monitoring_interval_seconds)
    logger.debug("Run monitor thread shutting down")


def run_worker_monitoring_thread_iteration(
    instance: DagsterCloudAgentInstance,
    statuses_list: List[CloudRunWorkerStatus],
    statuses_lock: threading.Lock,
    logger: logging.Logger,
):
    try:
        statuses = get_cloud_run_worker_statuses(instance)
        logger.debug("Thread got statuses: {}".format(statuses))
        with statuses_lock:
            statuses_list.clear()
            statuses_list.extend(statuses)
    except Exception:
        error_info = serializable_error_info_from_exc_info(sys.exc_info())
        logger.error("Caught error in run monitoring thread:\n{}".format(error_info))


def start_run_worker_monitoring_thread(
    instance: DagsterCloudAgentInstance,
    statuses_list: List[CloudRunWorkerStatus],
    statuses_lock: threading.Lock,
):
    shutdown_event = threading.Event()
    thread = threading.Thread(
        target=run_worker_monitoring_thread,
        args=(instance.get_ref(), statuses_list, statuses_lock, shutdown_event),
        name="run-monitoring",
    )
    thread.start()
    return thread, shutdown_event



from pathlib import Path

import requests
from typer import Argument, Option, Typer

from ....errors import raise_http_error
from ....headers.impl import get_dagster_cloud_api_headers
from ... import gql, ui
from ...config_utils import dagster_cloud_options

app = Typer(help="Customize your SAML settings.")


@app.command(name="upload-identity-provider-metadata")
@dagster_cloud_options(allow_empty=True)
def upload_identity_provider_metadata_command(
    api_token: str,
    organization: str,
    url: str = Option(
        None,
        "--url",
        help="[DEPRECATED] Your Dagster Cloud url, in the form of 'https://{ORGANIZATION_NAME}.dagster.cloud'.",
        hidden=True,
    ),
    metadata_file: Path = Argument(..., readable=True, metavar="METADATA_FILE_PATH"),
):
    """Upload your identity provider's metadata to enable SAML Single sign-on (SSO)."""

    if not url and not organization:
        raise ui.error("Must provide either organization name or URL.")

    if not url:
        url = gql.url_from_config(organization=organization)

    with open(metadata_file, "r", encoding="utf8") as f:
        response = requests.post(
            url=f"{url}/upload_idp_metadata",
            headers=get_dagster_cloud_api_headers(api_token),
            files={"metadata.xml": f},
        )

    raise_http_error(response)

    ui.print("The identity provider metadata was successfully uploaded.")



import json
import random
import string
import webbrowser
from enum import Enum
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Optional, Tuple, Type
from urllib import parse

from dagster.utils import find_free_port
from typer import Argument, Context, Option, Typer

from .. import gql, ui
from ..config_utils import (
    DagsterCloudCliConfig,
    available_deployment_names,
    dagster_cloud_options,
    read_config,
    write_config,
)

app = Typer(help="Configure the Dagster Cloud CLI.")


@app.command()
@dagster_cloud_options()
def set_deployment(
    ctx: Context, deployment: str = Argument(..., autocompletion=available_deployment_names)
):
    """Set the default deployment for CLI commands."""
    deployments = available_deployment_names(ctx=ctx)
    if not deployment or not deployment in deployments:
        raise ui.error(f"Deployment {ui.as_code(deployment)} not found")

    config = read_config()
    new_config = config._replace(default_deployment=deployment)
    write_config(new_config)

    ui.print(f"Default deployment changed to {ui.as_code(deployment)}")


@app.command()
def view(
    show_token: bool = Option(
        False, "--show-token", "-s", help="Whether to display the user token in plaintext."
    )
):
    """View the current CLI configuration."""
    config = read_config()
    if not show_token and config.user_token:
        config = config._replace(user_token=ui.censor_token(config.user_token))
    config_to_display = {k: v for k, v in config._asdict().items() if v is not None}
    ui.print_yaml(config_to_display)


app_configure = Typer()


def create_token_callback_handler(nonce: str) -> Type[BaseHTTPRequestHandler]:
    class TokenCallbackHandler(BaseHTTPRequestHandler):
        def _send_shared_headers(self):
            self.send_header("Access-Control-Allow-Origin", "*")
            self.send_header("Access-Control-Allow-Methods", "POST,OPTIONS")
            self.send_header("Access-Control-Allow-Headers", "Content-Type")

        def _error(self, code, message=None):
            self.send_error(code, message)
            self._send_shared_headers()
            self.end_headers()

        def do_OPTIONS(self):
            self.send_response(200)
            self._send_shared_headers()
            self.end_headers()

        def do_POST(self):
            url = parse.urlparse(self.path)

            if url.path != "/callback":
                self._error(404, "Page not found")
                return

            content = self.rfile.read(int(self.headers["Content-Length"]))
            body = json.loads(content)

            if not body.get("nonce") or body.get("nonce") != nonce:
                self._error(400, "Invalid nonce")
                return
            if not body.get("token"):
                self._error(400, "No user token provided")
                return
            if not body.get("organization"):
                self._error(400, "No organization provided")
                return

            organization = body.get("organization")
            token = body.get("token")

            self.send_response(200)
            self._send_shared_headers()
            response = b'{ "ok": true }'
            self.send_header("Content-length", str(len(response)))
            self.end_headers()

            self.wfile.write(response)

            # Pass back result values
            assert isinstance(self.server, TokenServer)
            self.server.set_result(organization=organization, token=token)
            self.server.shutdown()

        # Overridden so that the webserver doesn't log requests to the console
        # pylint: disable=unused-argument
        def log_request(self, code="-", size="-"):
            return

    return TokenCallbackHandler


class TokenServer(HTTPServer):
    organization: Optional[str] = None
    token: Optional[str] = None

    def __init__(self, host: Tuple[str, int], nonce: str):
        super().__init__(host, create_token_callback_handler(nonce))

    def shutdown(self):
        # Stop serving the token server
        # https://stackoverflow.com/a/36017741
        setattr(self, "_BaseServer__shutdown_request", True)

    def set_result(self, organization: str, token: str):
        self.organization = organization
        self.token = token

    def get_organization(self) -> Optional[str]:
        return self.organization

    def get_token(self) -> Optional[str]:
        return self.token


class SetupAuthMethod(Enum):
    WEB = "web"
    CLI = "cli"


def _settings_method_input(api_token: str):
    if api_token:
        choices = [
            ui.choice(SetupAuthMethod.CLI, "Authenticate using token or keep current settings"),
            ui.choice(SetupAuthMethod.WEB, "Authenticate in browser"),
        ]
    else:
        choices = [
            ui.choice(SetupAuthMethod.WEB, "Authenticate in browser"),
            ui.choice(SetupAuthMethod.CLI, "Authenticate using token"),
        ]
    return ui.list_input(
        "How would you like to authenticate the CLI?",
        choices,
    )


def _generate_nonce():
    return "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(8)
    )


def _setup(organization: str, deployment: str, api_token: str):
    setup_method = _settings_method_input(api_token)

    if setup_method == SetupAuthMethod.WEB:
        port = find_free_port()
        nonce = _generate_nonce()
        escaped = parse.quote(f"/cli-auth/{nonce}?port={port}")
        auth_url = f"https://dagster.cloud?next={escaped}"
        try:
            webbrowser.open(auth_url, new=0, autoraise=True)
            ui.print(
                f"Opening browser...\nIf a window does not open automatically, visit {auth_url} to finish authorization"
            )
        except webbrowser.Error as e:
            ui.warn(
                f"Error launching web browser: {str(e)}\n\nTo finish authorization, visit {auth_url}\n"
            )

        server = TokenServer(("localhost", port), nonce)
        server.serve_forever()
        new_org = server.get_organization()
        new_api_token = server.get_token()
        ui.print(f"Authorized for organization {ui.as_code(str(new_org))}\n")
    else:
        new_org = ui.input("Dagster Cloud organization:", default=organization or "") or None
        new_api_token = (
            ui.password_input("Dagster Cloud user token:", default=api_token or "") or None
        )

    # Attempt to fetch deployment names from server, fallback to a text input upon failure
    deployment_names = []
    if new_org and new_api_token:
        try:
            deployments = gql.fetch_deployments(
                gql.graphql_client_from_url(gql.url_from_config(new_org), new_api_token)
            )
            deployment_names = [deployment["deploymentName"] for deployment in deployments]
        except:
            ui.warn(
                "Could not fetch deployment names from server - organization or user token may be set incorrectly."
            )

    new_deployment: Optional[str] = None
    if deployment_names:
        options = ["None"] + deployment_names
        new_deployment = ui.list_input(
            "Default deployment:",
            choices=options,
            default=deployment if deployment in options else "None",
        )
        if new_deployment == "None":
            new_deployment = None
    else:
        new_deployment = ui.input("Default deployment:", default=deployment or "") or None

    write_config(
        DagsterCloudCliConfig(
            organization=new_org, default_deployment=new_deployment, user_token=new_api_token
        )
    )


@app.command()
@dagster_cloud_options(allow_empty=True)
def setup(organization: str, deployment: str, api_token: str):
    """Populate the CLI configuration."""
    _setup(organization, deployment, api_token)


app_configure = Typer(hidden=True)

# Legacy, to support the old `dagster-cloud configure` path
# New command is `dagster-cloud config setup`
@app_configure.command(name="configure", hidden=True)
@dagster_cloud_options(allow_empty=True)
def configure_legacy(organization: str, deployment: str, api_token: str):
    """Populate the CLI configuration."""
    _setup(organization, deployment, api_token)



import sys

from dagster import PipelineRun
from dagster.core.launcher.base import CheckRunHealthResult, WorkerStatus
from dagster.utils.error import serializable_error_info_from_exc_info
from dagster_k8s import K8sRunLauncher
from dagster_k8s.job import get_job_name_from_run_id


class CloudK8sRunLauncher(K8sRunLauncher):
    # Fork to avoid call to `count_resume_run_attempts`, since resuming runs is not currently
    # supported in cloud and the method makes repeated event log calls.
    def check_run_worker_health(self, run: PipelineRun):

        container_context = self.get_container_context_for_run(run)

        job_name = get_job_name_from_run_id(
            run.run_id,
        )
        try:
            job = self._batch_api.read_namespaced_job(
                namespace=container_context.namespace, name=job_name
            )
        except Exception:
            return CheckRunHealthResult(
                WorkerStatus.UNKNOWN, str(serializable_error_info_from_exc_info(sys.exc_info()))
            )
        if job.status.failed:
            return CheckRunHealthResult(WorkerStatus.FAILED, "K8s job failed")
        if job.status.succeeded:
            return CheckRunHealthResult(WorkerStatus.SUCCESS)
        return CheckRunHealthResult(WorkerStatus.RUNNING)
